class Acting
  def initialize(cast)
    @actors = cast.map do |actor, role|
      Actor.new(actor, role)
    end
  end

  def play(&block)
    if block_given?
      play_with_block(&block)
    else
      @actors.each(&:play)
    end
  end

  def quit
    @actors.each(&:quit)
  end

  private
  def play_with_block(&block)
    @actors.each(&:play)

    block.call
  ensure
    quit
  end

  module Cast
    private
    def __inject_methods__(methods)
      methods.each do |method|
        __injected_methods__[method.name] << method
      end
    end

    def __remove_methods__(methods)
      methods.each do |method|
        __injected_methods__[method.name].pop
      end
    end

    def __injected_methods__
      @__injected_methods__ ||= Hash.new do |hash, key|
        hash[key] = []
      end
    end

    def method_missing(meth, *args, &blk)
      method = __injected_methods__[meth].last
      if method
        method.bind(self).call(*args)
      else
        super
      end
    end
  end

  class Actor
    def initialize(actor, role)
      raise "#{actor.inspect} not kind of Acting::Cast" unless actor.kind_of? Cast
      @actor = actor
      @role = role
    end

    def play
      assign_behaviour
    end

    def quit
      revoke_behaviour
    end

    private
    def role_methods
      unless @role_methods
        method_names = @role.instance_methods

        if RUBY_VERSION =~ /2\.0/
          @role_methods = method_names.map do |name|
            @role.instance_method name
          end
        else
          # interim object so we can bind the method
          # willy nilly (no 'kind_of?(@role) == true' required)
          object = Object.new
          object.extend @role

          @role_methods = method_names.map do |name|
            object.method(name).unbind
          end
        end

      end
      @role_methods
    end

    def assign_behaviour
      @actor.send :__inject_methods__, role_methods
    end

    def revoke_behaviour
      @actor.send :__remove_methods__, role_methods
    end
  end
end
